@@ -1,16 +1,3 @@
-"""
-Copyright 2015 Zalando SE
-
-Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
-License. You may obtain a copy of the License at
-
-http://www.apache.org/licenses/LICENSE-2.0
-
-Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
-"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
- language governing permissions and limitations under the License.
-"""
-
 import functools
 import logging
 from copy import deepcopy
@@ -18,22 +5,34 @@ from copy import deepcopy
 from jsonschema import ValidationError
 
 from .decorators import validation
+from .decorators.decorator import (BeginOfRequestLifecycleDecorator,
+                                   EndOfRequestLifecycleDecorator)
 from .decorators.metrics import UWSGIMetricsCollector
 from .decorators.parameter import parameter_to_arg
-from .decorators.produces import BaseSerializer, Jsonifier, Produces
+from .decorators.produces import BaseSerializer, Produces
 from .decorators.response import ResponseValidator
 from .decorators.security import (get_tokeninfo_url, security_passthrough,
                                   verify_oauth)
 from .decorators.validation import (ParameterValidator, RequestBodyValidator,
                                     TypeValidationError)
 from .exceptions import InvalidSpecification
-from .utils import flaskify_endpoint, is_nullable, produces_json
+from .utils import all_json, is_nullable
 
 logger = logging.getLogger('connexion.operation')
 
+DEFAULT_MIMETYPE = 'application/json'
+
+
+VALIDATOR_MAP = {
+    'parameter': ParameterValidator,
+    'body': RequestBodyValidator,
+    'response': ResponseValidator,
+}
+
 
 class SecureOperation(object):
-    def __init__(self, security, security_definitions):
+
+    def __init__(self, api, security, security_definitions):
         """
         :param security: list of security rules the application uses by default
         :type security: list
@@ -41,6 +40,7 @@ class SecureOperation(object):
             <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
         :type security_definitions: dict
         """
+        self.api = api
         self.security = security
         self.security_definitions = security_definitions
 
@@ -97,18 +97,43 @@ class SecureOperation(object):
         # if we don't know how to handle the security or it's not defined we will usa a passthrough decorator
         return security_passthrough
 
+    def get_mimetype(self):
+        return DEFAULT_MIMETYPE
+
+    @property
+    def _request_begin_lifecycle_decorator(self):
+        """
+        Transforms the result of the operation handler in a internal
+        representation (connexion.decorators.ConnexionRequest) to be
+        used by internal Connexion decorators.
+
+        :rtype: types.FunctionType
+        """
+        return BeginOfRequestLifecycleDecorator(self.api, self.get_mimetype())
+
+    @property
+    def _request_end_lifecycle_decorator(self):
+        """
+        Guarantees that instead of the internal representation of the
+        operation handler response
+        (connexion.decorators.ConnexionRequest) a flask.Response
+        object is returned.
+        :rtype: types.FunctionType
+        """
+        return EndOfRequestLifecycleDecorator(self.api, self.get_mimetype())
+
 
 class Operation(SecureOperation):
-    DEFAULT_MIMETYPE = 'application/json'
 
     """
     A single API operation on a path.
     """
 
-    def __init__(self, method, path, operation, resolver, app_produces,
+    def __init__(self, api, method, path, operation, resolver, app_produces, app_consumes,
                  path_parameters=None, app_security=None, security_definitions=None,
                  definitions=None, parameter_definitions=None, response_definitions=None,
-                 validate_responses=False, strict_validation=False):
+                 validate_responses=False, strict_validation=False, randomize_endpoint=None,
+                 validator_map=None, pythonic_params=False):
         """
         This class uses the OperationID identify the module and function that will handle the operation
 
@@ -128,6 +153,10 @@ class Operation(SecureOperation):
         :param resolver: Callable that maps operationID to a function
         :param app_produces: list of content types the application can return by default
         :type app_produces: list
+        :param app_consumes: list of content types the application consumes by default
+        :type app_consumes: list
+        :param validator_map: map of validators
+        :type validator_map: dict
         :param path_parameters: Parameters defined in the path level
         :type path_parameters: list
         :param app_security: list of security rules the application uses by default
@@ -142,14 +171,22 @@ class Operation(SecureOperation):
         :type parameter_definitions: dict
         :param response_definitions: Global response definitions
         :type response_definitions: dict
+        :param validator_map: Custom validators for the types "parameter", "body" and "response".
+        :type validator_map: dict
         :param validate_responses: True enables validation. Validation errors generate HTTP 500 responses.
         :type validate_responses: bool
         :param strict_validation: True enables validation on invalid request parameters
         :type strict_validation: bool
+        :param pythonic_params: When True CamelCase parameters are converted to snake_case and an underscore is appended
+        to any shadowed built-ins
+        :type pythonic_params: bool
         """
 
+        self.api = api
         self.method = method
         self.path = path
+        self.validator_map = dict(VALIDATOR_MAP)
+        self.validator_map.update(validator_map or {})
         self.security_definitions = security_definitions or {}
         self.definitions = definitions or {}
         self.parameter_definitions = parameter_definitions or {}
@@ -162,6 +199,8 @@ class Operation(SecureOperation):
         self.validate_responses = validate_responses
         self.strict_validation = strict_validation
         self.operation = operation
+        self.randomize_endpoint = randomize_endpoint
+        self.pythonic_params = pythonic_params
 
         # todo support definition references
         # todo support references to application level parameters
@@ -171,10 +210,10 @@ class Operation(SecureOperation):
 
         self.security = operation.get('security', app_security)
         self.produces = operation.get('produces', app_produces)
+        self.consumes = operation.get('consumes', app_consumes)
 
         resolution = resolver.resolve(self)
         self.operation_id = resolution.operation_id
-        self.endpoint_name = flaskify_endpoint(self.operation_id)
         self.__undecorated_function = resolution.function
 
         self.validate_defaults()
@@ -275,15 +314,15 @@ class Operation(SecureOperation):
 
         :rtype str
         """
-        if produces_json(self.produces):
+        if all_json(self.produces):
             try:
                 return self.produces[0]
             except IndexError:
-                return Operation.DEFAULT_MIMETYPE
+                return DEFAULT_MIMETYPE
         elif len(self.produces) == 1:
             return self.produces[0]
         else:
-            return Operation.DEFAULT_MIMETYPE
+            return DEFAULT_MIMETYPE
 
     def resolve_parameters(self, parameters):
         for param in parameters:
@@ -291,7 +330,8 @@ class Operation(SecureOperation):
             yield param
 
     def get_path_parameter_types(self):
-        return {p['name']: p.get('type') for p in self.parameters if p['in'] == 'path'}
+        return {p['name']: 'path' if p.get('type') == 'string' and p.get('format') == 'path' else p.get('type')
+                for p in self.parameters if p['in'] == 'path'}
 
     @property
     def body_schema(self):
@@ -324,7 +364,9 @@ class Operation(SecureOperation):
         :rtype: types.FunctionType
         """
 
-        function = parameter_to_arg(self.parameters, self.__undecorated_function)
+        function = parameter_to_arg(
+            self.parameters, self.consumes, self.__undecorated_function, self.pythonic_params)
+        function = self._request_begin_lifecycle_decorator(function)
 
         if self.validate_responses:
             logger.debug('... Response validation enabled.')
@@ -333,7 +375,7 @@ class Operation(SecureOperation):
             function = response_decorator(function)
 
         produces_decorator = self.__content_type_decorator
-        logger.debug('... Adding produces decorator (%r)', produces_decorator, extra=vars(self))
+        logger.debug('... Adding produces decorator (%r)', produces_decorator)
         function = produces_decorator(function)
 
         for validation_decorator in self.__validation_decorators:
@@ -341,13 +383,15 @@ class Operation(SecureOperation):
 
         # NOTE: the security decorator should be applied last to check auth before anything else :-)
         security_decorator = self.security_decorator
-        logger.debug('... Adding security decorator (%r)', security_decorator, extra=vars(self))
+        logger.debug('... Adding security decorator (%r)', security_decorator)
         function = security_decorator(function)
 
         if UWSGIMetricsCollector.is_available():  # pragma: no cover
             decorator = UWSGIMetricsCollector(self.path, self.method)
             function = decorator(function)
 
+        function = self._request_end_lifecycle_decorator(function)
+
         return function
 
     @property
@@ -370,14 +414,16 @@ class Operation(SecureOperation):
         logger.debug('... Produces: %s', self.produces, extra=vars(self))
 
         mimetype = self.get_mimetype()
-        if produces_json(self.produces):  # endpoint will return json
+        if all_json(self.produces):  # endpoint will return json
             logger.debug('... Produces json', extra=vars(self))
-            jsonify = Jsonifier(mimetype)
+            jsonify = self.api.jsonifier(mimetype)
             return jsonify
+
         elif len(self.produces) == 1:
             logger.debug('... Produces %s', mimetype, extra=vars(self))
             decorator = Produces(mimetype)
             return decorator
+
         else:
             return BaseSerializer()
 
@@ -386,10 +432,14 @@ class Operation(SecureOperation):
         """
         :rtype: types.FunctionType
         """
+        ParameterValidator = self.validator_map['parameter']
+        RequestBodyValidator = self.validator_map['body']
         if self.parameters:
-            yield ParameterValidator(self.parameters, strict_validation=self.strict_validation)
+            yield ParameterValidator(self.parameters,
+                                     self.api,
+                                     strict_validation=self.strict_validation)
         if self.body_schema:
-            yield RequestBodyValidator(self.body_schema,
+            yield RequestBodyValidator(self.body_schema, self.consumes, self.api,
                                        is_nullable(self.body_definition))
 
     @property
@@ -398,4 +448,13 @@ class Operation(SecureOperation):
         Get a decorator for validating the generated Response.
         :rtype: types.FunctionType
         """
+        ResponseValidator = self.validator_map['response']
         return ResponseValidator(self, self.get_mimetype())
+
+    def json_loads(self, data):
+        """
+        A Wrapper for calling the jsonifier.
+        :param data: The json to loads
+        :type data: bytes
+        """
+        return self.api.jsonifier.loads(data)
